export MomentumRotorParameters, MomentumRotorControls, MomentumRotor, MomentumRotorStates
export construct_momentum_rotor_10mw, thrust_coefficient, power_coefficient
export update_controls!, calculate_thrust, calculate_aerodynamic_power
export step_forward!

abstract type RotorParameters end
abstract type Rotor end
abstract type RotorControls end
abstract type RotorStates end

@kwdef struct MomentumRotorParameters <: RotorParameters
    diameter
    time_step
    radius = diameter/2
    area = π*radius^2
    air_density = 1.225
end

@kwdef mutable struct MomentumRotorControls <: RotorControls
    induction
    velocity
    omega=0
end
@kwdef mutable struct MomentumRotorStates <: RotorStates
    azimuth
end

@kwdef struct MomentumRotor <: Rotor
    parameters::MomentumRotorParameters
    controls::MomentumRotorControls
    states::MomentumRotorStates
end

"""
    construct_momentum_rotor_10mw(h=0.3)

Construct a basic rotor struct for a DTU 10 MW reference turbine
"""
function construct_momentum_rotor_10mw(h=0.3)
   parameters = MomentumRotorParameters(diameter=178.3, time_step=h*17.83)
    controls = MomentumRotorControls(induction=0.33, velocity=SVector{3}([10.,0,0]))
    states = MomentumRotorStates(azimuth=0.)
    return MomentumRotor(parameters, controls, states)
end

"""
    thrust_coefficient(rotor::MomentumRotor)

Compute thrust coefficient for `rotor` object.
"""
function thrust_coefficient(rotor::MomentumRotor)
    a = rotor.controls.induction
    ct1 = 2.3
    a_t = 1-0.5*sqrt(ct1)
    if a > a_t
        ctp = ct1 / (1-a)^2 - 4 * (sqrt(ct1)-1) / (1-a)
    else
        ctp = 4a / (1-a)
    end
    return ctp
end

"""
    power_coefficient(rotor::MomentumRotor)

Compute power coefficient for `rotor` object
"""
function power_coefficient(rotor::MomentumRotor)
    a = rotor.controls.induction
    cp = 4*a / (1-a)
    return cp
end

"""
    calculate_thrust(rotor::Rotor, tower::Tower)

Calculate thrust by `rotor`, aligned according to `tower` orientation.
"""
function calculate_thrust(rotor::Rotor, tower::Tower)
    a = rotor.controls.induction
    ct = thrust_coefficient(rotor) * (1-a)^2
    A = rotor.parameters.area
    udotn = dot(rotor.controls.velocity, rotor_direction(tower))
    umag = norm(rotor.controls.velocity)
    thrust = 0.5 * rotor.parameters.air_density * A * ct * umag^2 * udotn^1
    return thrust
end

"""
    calculate_aerodynamic_power(rotor::Rotor,tower::Tower)

Calculate aerodynamic power at the rotor position
"""
function calculate_aerodynamic_power(rotor::Rotor,tower::Tower)
    cp = power_coefficient(rotor)
    A = rotor.parameters.area
    udotn = dot(rotor.controls.velocity, rotor_direction(tower))
    umag = norm(rotor.controls.velocity)
    thrust = 0.5 * rotor.parameters.air_density * A * cp * umag^3 * udotn^3
    return thrust
end

"""
    update_controls!(rotor::MomentumRotor, controls)

Update controls of `rotor` object with induction factor, free-stream velocity, and angular velocity
"""
function update_controls!(rotor::MomentumRotor, controls)
    rotor.controls.induction = controls[1]
    rotor.controls.velocity = controls[2]
    rotor.controls.omega = controls[3]
 end

"""
    transform(points, rotor::Rotor)

Transform points in rotor frame according to azimuth angle
"""
function transform(points, rotor::Rotor)
    return map(p -> rot_x(rotor.states.azimuth) * p, points)    
end

"""
    step_forward!(rotor::MomentumRotor)

Evolve azimuthal angle of `rotor` in time.
"""
function step_forward!(rotor::MomentumRotor)
    rotor.states.azimuth += rotor.parameters.time_step * rotor.controls.omega
end