export generate_default_controls, generate_inflow
export run_forward!

"""
    generate_default_controls(num_steps)

Generate a vector of default controls for wake simulation (a=0.33, yaw=0)
"""
function generate_default_controls(num_steps)
    return [SVector{2}([0.33, 0.]) for k in 0:1:num_steps]
end

"""
    generate__inflow(speed, direction)

Generate inflow velocity vector with `speed` and `direction` in degrees
"""
function generate_inflow(speed, direction)
    return rot_z(-deg2rad(direction)) * SVector{3}([speed, 0., 0.])
end

"""
    run_forward!(wake, controls, inflow)

Evolve `wake`` over time with given `controls`` and `inflow velocity``
"""
function run_forward!(wake, controls, inflow)
    omega = 6. * norm(inflow)/wake.parameters.rotor.parameters.radius

    for k in eachindex(controls)
        # @show controls[k]
        induction, yaw = controls[k]

        u_rotor = rotor_velocity(wake)
        thrust = calculate_thrust(wake.parameters.rotor, wake.parameters.tower)

        # todo: find a better way to implement this, too easy to make mistakes here
        update_controls!(wake, [thrust, inflow..., omega])
        update_controls!(wake.parameters.rotor, [induction,inflow, omega] )
        update_controls!(wake.parameters.tower, [thrust, yaw] )

        step_forward!(wake)
        step_forward!(wake.parameters.rotor)
        step_forward!(wake.parameters.tower)

    end
end

"""
    run_forward_substep!(wake, controls, inflow,substep=4)

Evolve `wake`` over time with given `controls`` and `inflow velocity``, `substep` between rings
"""
function run_forward_substep!(wake, controls, inflow,substep=4)
    omega = 6. * norm(inflow)/wake.parameters.rotor.parameters.radius

    for k in eachindex(controls)
        # @show controls[k]
        induction, yaw = controls[k]

        u_rotor = rotor_velocity(wake)
        thrust = calculate_thrust(wake.parameters.rotor)

        # todo: find a better way to implement this, too easy to make mistakes here
        update_controls!(wake, [thrust, inflow..., omega])
        update_controls!(wake.parameters.rotor, [induction,inflow, omega] )
        update_controls!(wake.parameters.tower, [thrust, yaw] )

        step_forward_substep!(wake,substep)
        step_forward!(wake.parameters.rotor)
        step_forward!(wake.parameters.tower)

    end
end
"""
    run_forward_Heun!(wake, controls, inflow,substep=4)

Evolve `wake`` over time with given `controls`` and `inflow velocity``, `substep` between rings

Heun's method instead of explicit Euler (2nd instead of 1st order convergence over time)
"""
function run_forward_Heun!(wake::Union{PlanarWake,ADMWake}, controls, inflow,substep=1)
    omega = 6. * norm(inflow)/wake.parameters.rotor.parameters.radius

    for k in eachindex(controls)
        # @show controls[k]
        induction, yaw = controls[k]

        @show u_rotor = rotor_velocity(wake)
        thrust = calculate_thrust(wake.parameters.rotor)

        # todo: find a better way to implement this, too easy to make mistakes here
        update_controls!(wake, [thrust, inflow..., omega])
        update_controls!(wake.parameters.rotor, [induction,inflow, omega] )
        update_controls!(wake.parameters.tower, [thrust, yaw] )

        step_forward_Heun!(wake,substep)
        step_forward!(wake.parameters.rotor)
        step_forward!(wake.parameters.tower)

    end
end
