export FixedTowerParameters, FixedTowerStates, FixedTowerControls, FixedTower
export construct_fixed_tower_10mw
export update_controls!, step_forward!
export transform, motion, rotor_direction


abstract type TowerParameters end
abstract type TowerStates end
abstract type TowerControls end
abstract type Tower end


@kwdef struct FixedTowerParameters <: TowerParameters
    base_position
    hubheight
    tilt
    # time_step
end

@kwdef mutable struct FixedTowerStates <: TowerStates
    base_pitch=0.
    base_pitch_rate=0.
    base_surge=0.
    base_surge_rate=0.
end

@kwdef mutable struct FixedTowerControls <: TowerControls
    thrust
    yaw
end

@kwdef mutable struct FixedTower <: Tower
    parameters::FixedTowerParameters
    states::FixedTowerStates
    controls::FixedTowerControls
end
"""
    update_controls!(tower::FixedTower, controls)

Update tower controls with `thrust` and `yaw`
"""
function update_controls!(tower::FixedTower, controls)
    tower.controls.thrust = controls[1]
    tower.controls.yaw = controls[2]
end

"""
    step_forward!(tower::Tower)

Evolve tower dynamics in time
"""
function step_forward!(tower::Tower)
end

"""
    construct_fixed_tower_10mw(position)

Construct `tower` object for standard DTU 10 MW reference
"""
function construct_fixed_tower_10mw(position)
    params = FixedTowerParameters(position, 119., 0., )
    states = FixedTowerStates()
    controls = FixedTowerControls(0., 0.)
    return FixedTower(params, states, controls)
end

"""
    transform_point(point::SVector, tower::Tower)

Transform point from nacelle centre coordinate to global coordinates
"""
function transform_point(point::SVector{2}, tower::Tower)
    return rot_z(deg2rad(-tower.controls.yaw))[1:2,1:2] * rot_y(-deg2rad(tower.parameters.tilt))[1:2,1:2] * point +
        tower.parameters.base_position[1:2] +
        (rot_y(deg2rad(tower.states.base_pitch)) * SVector{3}([0., 0., tower.parameters.hubheight]))[1:2] +
        SVector{2}([tower.states.base_surge, 0.])
end
function transform_point(point::SVector{3}, tower::Tower)
    return rot_z(deg2rad(-tower.controls.yaw)) * rot_y(-deg2rad(tower.parameters.tilt)) * point +
        tower.parameters.base_position +
        rot_y(deg2rad(tower.states.base_pitch)) * SVector{3}([0., 0., tower.parameters.hubheight]) +
        SVector{3}([tower.states.base_surge, 0., 0.])
end

"""
    transform(points, tower::Tower)

Transform list of points from nacelle centre coordinate to global coordinates
"""
function transform(points, tower::Tower)
    return map(p -> transform_point(p, tower), points)
end

"""
    motion_point(point::SVector{2}, tower::FixedTower)

Calculate motion of point from tower dynamics
"""
function motion_point(point::SVector{2}, tower::FixedTower)
    return @SVector zeros(2)
end
function motion_point(point::SVector{3}, tower::FixedTower)
    return @SVector zeros(3)
end

"""
    motion(points, tower::Tower)

Caculate motion of list of points from nacelle centre coordinate in global reference
"""
function motion(points, tower::Tower)
    return map(p -> motion_point(p, tower), points)
end

"""
    rotor_direction(tower::Tower)

Set up vector normal to rotor
"""
function rotor_direction(tower::Tower)
    return rot_z(deg2rad(-tower.controls.yaw)) * rot_y(-deg2rad(tower.parameters.tilt))* SVector{3}([1., 0., 0.])
end

"""
    rot_x(phi)

Construct rotation matrix around x axis, angle in radians
"""
function rot_x(phi)
    R = SMatrix{3,3}([1.  0.    0.;
        0.  cos(phi)   -sin(phi);
        0.  sin(phi)     cos(phi)])
    return R
end

"""
    rot_y(phi)

Construct rotation matrix around y axis, angle in radians
"""
function rot_y(phi)
    R = SMatrix{3,3}([cos(phi) 0.    sin(phi);
        0.         1.    0.;
        -sin(phi)  0.     cos(phi)])
    return R
end

"""
    rot_z(phi)

Construct rotation matrix around z axis, angle in radians
"""
function rot_z(phi)
    R = SMatrix{3,3}([cos(phi)  -sin(phi)     0;
        sin(phi)  cos(phi)    0.;
        0.        0.     1.])
    return R
end