export ADMWakeStates, ADMWakeParameters, ADMWakeControls, ADMWake
export ADMRWakeStates, ADMRWakeParameters, ADMRWakeControls, ADMRWake
export JoukowskyWakeStates, JoukowskyWakeParameters, JoukowskyWakeControls, JoukowskyWake
export PlanarWakeStates, PlanarWakeParameters, PlanarWakeControls, PlanarWake

export initialise_states, construct_default_adm_wake
export induced_velocity, velocity, rotor_velocity
export update_controls!, step_forward!
export rotor_disc_points
export construct_default_adm_wake
export construct_default_admr_wake
export construct_default_joukowsky_wake

abstract type Wake end
abstract type WakeParameters end
abstract type WakeStates end
abstract type WakeControls end

@kwdef struct PlanarWakeParameters <:WakeParameters
    rotor::Rotor
    tower::Tower
    num_rings::Int
    num_elements::Int
    rotor_points
    time_step
    initial_core_size
    alpha=1.25643
    delta=0.
    kinematic_viscosity=1.5e-5
end
@kwdef mutable struct PlanarWakeStates <: WakeStates
    positions
    vortex_strengths
    core_sizes
end
@kwdef mutable struct PlanarWakeControls <: WakeControls
    thrust
    inflow
end
@kwdef struct PlanarWake <: Wake
    parameters::PlanarWakeParameters
    states::PlanarWakeStates
    controls::PlanarWakeControls
end
@kwdef struct ADMWakeParameters <:WakeParameters
    rotor::Rotor
    tower::Tower
    num_rings::Int
    num_elements::Int
    rotor_points
    time_step
    initial_core_size
    alpha=1.25643
    delta=0.
    kinematic_viscosity=1.5e-5
    mirror_ground=false
end
@kwdef mutable struct ADMWakeStates <: WakeStates
    positions
    vortex_strengths
    core_sizes
end
@kwdef mutable struct ADMWakeControls <: WakeControls
    thrust
    inflow
end
@kwdef struct ADMWake <: Wake
    parameters::ADMWakeParameters
    states::ADMWakeStates
    controls::ADMWakeControls
end
@kwdef struct ADMRWakeParameters <:WakeParameters
    rotor::Rotor
    tower::Tower
    num_rings::Int
    num_elements::Int
    rotor_points
    time_step
    initial_core_size
    alpha=1.25643
    delta=0.
    kinematic_viscosity=1.5e-5
    mirror_ground=false
end
@kwdef mutable struct ADMRWakeStates <: WakeStates
    positions
    vortex_strengths
    core_sizes
end
@kwdef mutable struct ADMRWakeControls <: WakeControls
    thrust
    inflow
    omega
end
@kwdef struct ADMRWake <: Wake
    parameters::ADMRWakeParameters
    states::ADMRWakeStates
    controls::ADMRWakeControls
end

@kwdef struct JoukowskyWakeParameters <:WakeParameters
    rotor::Rotor
    tower::Tower
    num_rings::Int
    num_blades::Int
    rotor_points
    time_step
    initial_core_size
    alpha=1.25643
    delta=0.
    kinematic_viscosity=1.5e-5
end
@kwdef mutable struct JoukowskyWakeStates <: WakeStates
    positions
    vortex_strengths
    core_sizes
end
@kwdef mutable struct JoukowskyWakeControls <: WakeControls
    thrust
    inflow
    omega
end
@kwdef struct JoukowskyWake <: Wake
    parameters::JoukowskyWakeParameters
    states::    JoukowskyWakeStates
    controls::  JoukowskyWakeControls
end


"""
    rotor_disc_points()

Generate list of points for sampling rotor plane velocities
"""
function rotor_disc_points()
    N1 = 3
    nrings = 3
    i = 1:1:nrings
    Ni = (2 .* i .- 1) .* N1
    p = zeros((sum(Ni), 3))
    Ri = (i .- 0.5) .* (0.5 / nrings)
    N0 = 0
    for r in 1:1:nrings
        Thi = range(start=0, stop=2 * π - 2* π / Ni[r] , step= 2* π / Ni[r])
        p[(N0+1:N0 + Ni[r]), 2] = Ri[r] .* sin.(Thi)
        p[(N0+1:N0 + Ni[r]), 3] = Ri[r] .* cos.(Thi)
        N0 += Ni[r]
    end
    return [SVector{3}(x) for x in eachrow(p)]
end

"""
    construct_default_adm_wake()

Construct a basic ADMWake for modelling a DTU 10 MW reference
"""
function construct_default_adm_wake()
    rotor = construct_momentum_rotor_10mw()
    tower = construct_fixed_tower_10mw(@SVector zeros(3))
    params = ADMWakeParameters(rotor=rotor, tower=tower,
        num_rings=40, num_elements=16,
        rotor_points=rotor_disc_points() * rotor.parameters.diameter,
        time_step=0.3 * rotor.parameters.diameter / 10.,
         initial_core_size=0.16 * rotor.parameters.diameter)
    states = initialise_states(params)
    controls = ADMWakeControls(thrust = 0., inflow = SVector{3}(zeros(3)))
    return ADMWake(params, states, controls)
end

"""
    construct_default_admr_wake()

Construct a basic ADMRWake for modelling a DTU 10 MW reference
"""
function construct_default_admr_wake()
    rotor = construct_momentum_rotor_10mw()
    tower = construct_fixed_tower_10mw(@SVector zeros(3))
    params = ADMRWakeParameters(rotor=rotor, tower=tower,
        num_rings=40, num_elements=16,
        rotor_points=rotor_disc_points() * rotor.parameters.diameter,
        time_step=0.3 * rotor.parameters.diameter / 10.,
         initial_core_size=0.16 * rotor.parameters.diameter)
    states = initialise_states(params)
    controls = ADMRWakeControls(thrust = 0., inflow = SVector{3}(zeros(3)), omega=0.)
    return ADMRWake(params, states, controls)
end

"""
    construct_default_joukowsky_wake()

Construct a basic JoukowskyWake for modelling a DTU 10 MW reference
"""
function construct_default_joukowsky_wake()
    rotor = construct_momentum_rotor_10mw(0.033)
    tower = construct_fixed_tower_10mw(@SVector zeros(3))
    params = JoukowskyWakeParameters(rotor=rotor, tower=tower,
        num_rings=100, num_blades=3,
        rotor_points=rotor_disc_points() * rotor.parameters.diameter,
        time_step=0.033 * rotor.parameters.diameter / 10.,
         initial_core_size=  0.5*0.16 * rotor.parameters.diameter)
    states = initialise_states(params)
    controls = JoukowskyWakeControls(thrust = 0., inflow = SVector{3}(zeros(3)), omega=1.)
    return JoukowskyWake(params, states, controls)
end

"""
    initialise_states(p)

Initialise wake parameters to default settings
"""
function initialise_states(p::PlanarWakeParameters)
    X = [@SVector zeros(2) for e in 1:2, r in 1:p.num_rings]
    G = zeros(2, p.num_rings)  #.- 100.
    S = zeros(2, p.num_rings) .+ p.initial_core_size
    return PlanarWakeStates(positions=X,  vortex_strengths=G, core_sizes=S)
end
function initialise_states(p::ADMWakeParameters)
    X = [@SVector zeros(3) for e in 1:(p.num_elements), r in 1:p.num_rings]
    G = zeros(p.num_elements, p.num_rings)  #.- 100.
    S = zeros(p.num_elements, p.num_rings) .+ p.initial_core_size
    return ADMWakeStates(positions=X,  vortex_strengths=G, core_sizes=S)
end
function initialise_states(p::ADMRWakeParameters)
    # additional point at centre axis
    X = [@SVector zeros(3) for e in 1:(p.num_elements+1), r in 1:p.num_rings]
    G = zeros(p.num_elements+1, p.num_rings)  #.- 100.
    S = zeros(p.num_elements+1, p.num_rings) .+ p.initial_core_size
    return ADMRWakeStates(positions=X,  vortex_strengths=G, core_sizes=S)
end
function initialise_states(p::JoukowskyWakeParameters)
    # additional point at centre axis
    X = [SVector{3}([10. *r, 0,0]) for b in 1:(p.num_blades+1), r in 1:p.num_rings]
    G = zeros(2*p.num_blades+1, p.num_rings)  #.- 100.
    S = zeros(2*p.num_blades+1, p.num_rings) .+ p.initial_core_size
    # X = [SVector{3}([10. *r, 0,0]) for b in 1:(2*p.num_blades), r in 1:p.num_rings]
    # G = zeros(3*p.num_blades, p.num_rings)  #.- 100.
    # S = zeros(3*p.num_blades, p.num_rings) .+ p.initial_core_size #.* (3* [1.,1.,1.,10.,1.,1.,1.])
    return JoukowskyWakeStates(positions=X,  vortex_strengths=G, core_sizes=S)
end

"""
    initial_points(wake)

Construct points defining vortex elements shed from rotor edge
"""
function initial_points(wake::PlanarWake)
    # todo: test allocation of intermediate variable assignment
    p = wake.parameters
    pr = wake.parameters.rotor.parameters
    X0 = [SVector{2}([0, pr.radius*cos(phi)]) for phi in 
        [0, π]]
    return transform(X0, p.tower)
    # return transform(transform(X0, p.rotor), p.tower)
end
function initial_points(wake::ADMWake)
    # todo: test allocation of intermediate variable assignment
    p = wake.parameters
    pr = wake.parameters.rotor.parameters
    X0 = [SVector{3}([0, pr.radius*cos(phi), pr.radius*sin(phi)]) for phi in 
        range(start=0, step=2π/p.num_elements, length=p.num_elements)]
    return transform(X0, p.tower)
    # return transform(transform(X0, p.rotor), p.tower)
end
function initial_points(wake::ADMRWake)
    # todo: test allocation of intermediate variable assignment
    p = wake.parameters
    pr = wake.parameters.rotor.parameters
    X0 = [SVector{3}([0, pr.radius*cos(phi), pr.radius*sin(phi)]) for phi in 
        range(start=0, step=2π/p.num_elements, length=p.num_elements+1)]
    X0[end] = SVector{3}(zeros(3))
    # return transform(transform(X0, p.rotor), p.tower)
    return transform(X0, p.tower)
end
function initial_points(wake::JoukowskyWake)
    # todo: test allocation of intermediate variable assignment
    p = wake.parameters
    pr = wake.parameters.rotor.parameters
    # start = 0.1
    X0 = [SVector{3}([0, pr.radius*cos(phi), pr.radius*sin(phi)]) for phi in 
        range(start=0, step=2π/p.num_blades, length=p.num_blades+1)]
    # X0 = [SVector{3}([0, f*pr.radius*cos(phi), pr.radius*sin(phi)]) 
    #     for f in [start, 1] for phi in range(start=0, step=2π/p.num_blades, length=p.num_blades+1)]
    X0[end] = SVector{3}(zeros(3))
    return transform(transform(X0, p.rotor), p.tower)
end

"""
    induced_velocity(point, X, G, S)

Calculate induced velocity for point by element X with strength G and core S
"""
function induced_velocity(point::SVector{2}, X, G, S) 
    r0 = X .- [point]
    # filt = norm.(r0) .> -11e-8
    filt = norm.(r0) .> -1
    eps_reg = 1e-8
    # u = ([SMatrix{2,2}([0 1; -1 0])] .* r0[filt]).*
    # (G[filt] ./ (2π .* (norm.(r0[filt])).^2 )) .*
    # (1 .-exp.(-(norm.(r0).^2 ./(S.^2))))[filt]
    u = ([SMatrix{2,2}([0 1; -1 0])] .* r0[filt]).*
    (G[filt] ./ (2π .* (norm.(r0[filt])).^2 .+ eps_reg)) .*
    (1 .-exp.(-(norm.(r0).^2 ./(S.^2))))[filt]
    return sum(u)
end
"""
    induced_velocity(point, X1, X2, G, S)

Calculate induced velocity for element from X1 to X2 with strength G and core S
"""
function induced_velocity(point::SVector{3}, X1, X2, G, S) 
    r0 = X2 .- X1
    r1 = X1 .- [point] #[xv - point for xv in X1] # X1 .- point
    r2 = X2 .- [point] #[xv - point for xv in X2] #X2 .- point
    filt = norm.(cross.(r1, r2)) .>= 1e-8 .&& norm.(r1) .>= 1e-8 .&& norm.(r2) .> 1e-8
    u = ((G[filt] ./ 4π)  .* 
        cross.(r1[filt], r2[filt]) ./ (norm.(cross.(r1[filt], r2[filt]))).^2 .*
        dot.(r0[filt], (r1[filt]./norm.(r1[filt]) .- r2[filt]./norm.(r2[filt]))) .*  
        (1 .-exp.(-(norm.( cross.(r1, r2)).^2 ./(S.^2 .*norm.(r0).^2))))[filt]
        )
    return sum(u)
end
"""
    induced_velocity(points, X1, X2, G, S)

Calculate induced velocity for all points by elements  with strength G and core S
"""
function induced_velocity(points, X1, X2, G, S)
    return map(x -> induced_velocity(x, X1, X2, G, S), points)
end
function induced_velocity(points, X, G, S)
    return map(x -> induced_velocity(x, X, G, S), points)
end
function induced_velocity(points, wake::PlanarWake)
    return induced_velocity(points,
        wake.states.positions,
        wake.states.vortex_strengths,
        wake.states.core_sizes)
end
function induced_velocity(points, wake::ADMWake)
    if wake.parameters.mirror_ground
        x_mirror =  [x .* SVector{3}([1,1,-1]) for x in wake.states.positions]
        return induced_velocity(points,
        wake.states.positions,
        wake.states.positions[(1:end).%end.+1,:],
        wake.states.vortex_strengths,
        wake.states.core_sizes) .+ 
        induced_velocity(points,
        x_mirror,
        x_mirror[(1:end).%end.+1,:],
        wake.states.vortex_strengths.*-1,
        wake.states.core_sizes)
    else
        return induced_velocity(points,
        wake.states.positions,
        wake.states.positions[(1:end).%end.+1,:],
        wake.states.vortex_strengths,
        wake.states.core_sizes)
    end
end
function induced_velocity(points, wake::ADMRWake)
    p = wake.parameters
    if wake.parameters.mirror_ground
        x_mirror =  [x .* SVector{3}([1,1,-1]) for x in wake.states.positions]
        return induced_velocity(points,
        wake.states.positions[1:p.num_elements,:],
        wake.states.positions[(1:p.num_elements).%p.num_elements.+1,:],
        wake.states.vortex_strengths[1:p.num_elements,:],
        wake.states.core_sizes[1:p.num_elements,:]) .+
     induced_velocity(points,
        wake.states.positions[p.num_elements+1,1:p.num_rings-1],   
        wake.states.positions[p.num_elements+1,2:p.num_rings],
        wake.states.vortex_strengths[p.num_elements+1,1:p.num_rings-1],
        wake.states.core_sizes[p.num_elements+1,1:p.num_rings-1]) .+
        induced_velocity(points,
        x_mirror[1:p.num_elements,:],
        x_mirror[(1:p.num_elements).%p.num_elements.+1,:],
        wake.states.vortex_strengths[1:p.num_elements,:] .* -1,
        wake.states.core_sizes[1:p.num_elements,:]) .+
     induced_velocity(points,
        x_mirror[p.num_elements+1,1:p.num_rings-1],   
        x_mirror[p.num_elements+1,2:p.num_rings],
        wake.states.vortex_strengths[p.num_elements+1,1:p.num_rings-1] .* -1,
        wake.states.core_sizes[p.num_elements+1,1:p.num_rings-1])
    else
        return induced_velocity(points,
        wake.states.positions[1:p.num_elements,:],
        wake.states.positions[(1:p.num_elements).%p.num_elements.+1,:],
        wake.states.vortex_strengths[1:p.num_elements,:],
        wake.states.core_sizes[1:p.num_elements,:]) .+
     induced_velocity(points,
        wake.states.positions[p.num_elements+1,1:p.num_rings-1],   
        wake.states.positions[p.num_elements+1,2:p.num_rings],
        wake.states.vortex_strengths[p.num_elements+1,1:p.num_rings-1],
        wake.states.core_sizes[p.num_elements+1,1:p.num_rings-1])
    end
end
function induced_velocity(points, wake::JoukowskyWake)
    p = wake.parameters
    return induced_velocity(points, #induced velocity from blade bound and shed vorticity
        repeat(wake.states.positions[p.num_blades+1:p.num_blades+1,:], p.num_blades),
        wake.states.positions[1:p.num_blades,:],
        wake.states.vortex_strengths[p.num_blades+1 .+ (1:p.num_blades),:],
        wake.states.core_sizes[p.num_blades+1 .+ (1:p.num_blades),:]) .+
     induced_velocity(points, # induced velocity from tip and root vortex
        wake.states.positions[:,1:p.num_rings-1],   
        wake.states.positions[:,2:p.num_rings],       
        wake.states.vortex_strengths[1:p.num_blades+1,1:p.num_rings-1],
        wake.states.core_sizes[1:p.num_blades+1,1:p.num_rings-1])
end


"""
    velocity(points, wake::Wake)

Calculate total velocity from free-stream and induced at list of points
"""
function velocity(points, wake::Wake)
    return induced_velocity(points, wake) .+ [wake.controls.inflow]
end

"""
    rotor_velocity(wake::Wake)

Calculate rotor-effective velocity
"""
function rotor_velocity(wake::Wake)
    #todo:  not implemented
    p = transform(wake.parameters.rotor_points, wake.parameters.tower)
    return dot(rotor_direction(wake.parameters.tower), sum(velocity(p, wake) + motion(p, wake.parameters.tower))) / length(p)
end
function rotor_velocity(wake::PlanarWake)
    #todo:  not implemented
    p = transform(wake.parameters.rotor_points, wake.parameters.tower)
    return dot(rotor_direction(wake.parameters.tower)[1:2], sum(velocity(p, wake) + motion(p, wake.parameters.tower))) / length(p)
end

"""
    update_controls!(wake::PlanarWake, controls)

Update wake controls with thrust, inflow and rotor angular velocity
"""
function update_controls!(wake::PlanarWake, controls)
    wake.controls.thrust = controls[1]
    wake.controls.inflow = SVector{2}(controls[2:3])
end
function update_controls!(wake::ADMWake, controls)
    wake.controls.thrust = controls[1]
    wake.controls.inflow = SVector{3}(controls[2:4])
end
function update_controls!(wake::ADMRWake, controls)
    wake.controls.thrust = controls[1]
    wake.controls.inflow = SVector{3}(controls[2:4])
    wake.controls.omega = controls[5]
end
function update_controls!(wake::JoukowskyWake, controls)
    wake.controls.thrust = controls[1]
    wake.controls.inflow = SVector{3}(controls[2:4])
    wake.controls.omega = controls[5]
end
"""
    step_forward!(wake)

Evolve wake over one time-step, shedding and removing first and last set of points
"""
function step_forward!(wake::PlanarWake)
    #todo: not implemented
    X1 = Zygote.Buffer(wake.states.positions)
    G1 = Zygote.Buffer(wake.states.vortex_strengths)
    S1 = Zygote.Buffer(wake.states.core_sizes)
    p = wake.parameters
    q = wake.states
    X1[:,2:p.num_rings] = q.positions[:,1:p.num_rings-1] + 
        p.time_step * velocity(q.positions[:,1:p.num_rings-1], wake)
    G1[:,2:p.num_rings] = q.vortex_strengths[:,1:p.num_rings-1]
    # todo: implement vortex core size growth
    S1[:,2:p.num_rings] = q.core_sizes[:,1:p.num_rings-1]
    
    # generate new ring
    X1[:,1] = initial_points(wake)
    # uniform distribution of vortex strength
    gamma = p.time_step * wake.controls.thrust / (wake.parameters.rotor.parameters.air_density * p.rotor.parameters.area)
    G1[1,1] = -gamma 
    G1[2,1] = gamma
    # uniform core size initialisation
    S1[:,1] = p.initial_core_size * ones(p.num_elements)

    wake.states.positions = copy(X1)
    wake.states.vortex_strengths = copy(G1)
    wake.states.core_sizes = copy(S1)
end

function step_forward_substep!(wake::PlanarWake, substep=4)
    #todo: not implemented
    X1 = Zygote.Buffer(wake.states.positions)
    G1 = Zygote.Buffer(wake.states.vortex_strengths)
    S1 = Zygote.Buffer(wake.states.core_sizes)
    p = wake.parameters
    q = wake.states
    X1[:,2:p.num_rings] = q.positions[:,1:p.num_rings-1]+ 
    (p.time_step/substep) * velocity(q.positions[:,1:p.num_rings-1], wake)
    X1[:,1] = initial_points(wake)

    G1[:,2:p.num_rings] = q.vortex_strengths[:,1:p.num_rings-1]
    gamma = (1. /substep) * p.time_step * wake.controls.thrust / (wake.parameters.rotor.parameters.air_density * p.rotor.parameters.area)
    G1[1,1] = -gamma 
    G1[2,1] = gamma

    S1[:,2:p.num_rings] = q.core_sizes[:,1:p.num_rings-1]
    S1[:,1] = p.initial_core_size * ones(p.num_elements)
    wake.states.positions = copy(X1)
    wake.states.vortex_strengths = copy(G1)
    wake.states.core_sizes = copy(S1)

    for step in 2:1:substep
        X1 = Zygote.Buffer(wake.states.positions)
        G1 = Zygote.Buffer(wake.states.vortex_strengths)
        S1 = Zygote.Buffer(wake.states.core_sizes)

        X1[:,2:p.num_rings] = q.positions[:,2:p.num_rings]+ 
         (p.time_step/substep) * velocity(q.positions[:,2:p.num_rings], wake)
        X1[:,1] = initial_points(wake)
        G1[:] = wake.states.vortex_strengths
        gamma = (step/substep) * p.time_step * wake.controls.thrust / (wake.parameters.rotor.parameters.air_density * p.rotor.parameters.area)
        G1[1,1] = -gamma 
        G1[2,1] = gamma
        S1[:] = wake.states.core_sizes
        wake.states.positions = copy(X1)
        wake.states.vortex_strengths = copy(G1)
        wake.states.core_sizes = copy(S1)
    end
end

function step_forward_Heun!(wake::PlanarWake, substep=1)
    #todo: not implemented
    X1 = Zygote.Buffer(wake.states.positions)
    G1 = Zygote.Buffer(wake.states.vortex_strengths)
    S1 = Zygote.Buffer(wake.states.core_sizes)
    p = wake.parameters
    q = wake.states

    wake.states.vortex_strengths[:,end] .= 0
    # estimate state update with forward Euler
    velocity_k = velocity(q.positions[:,1:p.num_rings-1] , wake)
    X1_p = q.positions[:,1:p.num_rings-1] + (p.time_step/substep) * velocity_k
    
    
    w2 = deepcopy(wake)
    w2.states.positions[:,1] = initial_points(w2)
    w2.states.positions[:,2:end] = X1_p

    gamma = (p.time_step/substep) * wake.controls.thrust / (wake.parameters.rotor.parameters.air_density * p.rotor.parameters.area)
    w2.states.vortex_strengths[:,2:end] = w2.states.vortex_strengths[:,1:end-1]
    w2.states.vortex_strengths[1,1] = -gamma    
    w2.states.vortex_strengths[2,1] = gamma
    # w2.states.vortex_strengths[:,end] = (1-1/substep) .* w2.states.vortex_strengths[:,end]    
    # w2.states.vortex_strengths[2,1] = gamma
    w2.states.core_sizes[:,2:end] = w2.states.core_sizes[:,1:end-1]
    w2.states.core_sizes[:,1] .= p.initial_core_size
    velocity_k1 = velocity(X1_p , w2)

    X1[:,2:p.num_rings] = (1/2) *(p.time_step/substep)* (velocity_k .+ velocity_k1) + 
    q.positions[:,1:p.num_rings-1]
    G1[:,2:p.num_rings] = q.vortex_strengths[:,1:p.num_rings-1]
    # todo: implement vortex core size growth
    S1[:,2:p.num_rings] = q.core_sizes[:,1:p.num_rings-1]
    
    # generate new ring
    X1[:,1] = initial_points(wake)
    # uniform distribution of vortex strength
    gamma = (p.time_step/substep) * wake.controls.thrust / (wake.parameters.rotor.parameters.air_density * p.rotor.parameters.area)
    G1[1,1] = -gamma
    G1[2,1] = gamma
    # uniform core size initialisation
    S1[:,1] = p.initial_core_size * ones(p.num_elements)

    wake.states.positions = copy(X1)
    wake.states.vortex_strengths = copy(G1)
    wake.states.core_sizes = copy(S1)
    # wake.states.vortex_strengths[:,end] = (1-1/substep) .* wake.states.vortex_strengths[:,end]    

    for step in 2:1:substep
        X1 = Zygote.Buffer(wake.states.positions)
        G1 = Zygote.Buffer(wake.states.vortex_strengths)
        S1 = Zygote.Buffer(wake.states.core_sizes)
        p = wake.parameters
        q = wake.states
        velocity_k = velocity(q.positions[:,2:p.num_rings] , wake)
        X1_p = q.positions[:,2:p.num_rings] + (p.time_step/substep) * velocity_k
        
        w2 = deepcopy(wake)
        w2.states.positions[:,1] = initial_points(w2)
        w2.states.positions[:,2:end] = X1_p
        # w2.states.vortex_strengths[:,2:end] = w2.states.vortex_strengths[:,1:end-1]
        gamma = (step/substep) * p.time_step * wake.controls.thrust / (wake.parameters.rotor.parameters.air_density * p.rotor.parameters.area)
        w2.states.vortex_strengths[1,1] = -gamma
        w2.states.vortex_strengths[2,1] = gamma
        # w2.states.vortex_strengths[:,end] = ((1-step/substep)/(1-(step-1)/substep)) .* w2.states.vortex_strengths[:,end]    
        # w2.states.core_sizes[:,2:end] = w2.states.core_sizes[:,1:end-1]
        w2.states.core_sizes[:,1] .= p.initial_core_size
        velocity_k1 = velocity(X1_p , w2)
    
        X1[:,2:p.num_rings] = (1/2) *(p.time_step/substep)* (velocity_k + velocity_k1) + 
        q.positions[:,2:p.num_rings]

        # generate new ring
        X1[:,1] = initial_points(wake)
        G1[:] = wake.states.vortex_strengths
        gamma = (step/substep) * p.time_step * wake.controls.thrust / (wake.parameters.rotor.parameters.air_density * p.rotor.parameters.area)

        G1[1,1] = -gamma
        G1[2,1] = gamma

        S1[:] = wake.states.core_sizes
        wake.states.positions = copy(X1)
        wake.states.vortex_strengths = copy(G1)
        wake.states.core_sizes = copy(S1)
        # wake.states.vortex_strengths[:,end] = ((1-step/substep)/(1-(step-1)/substep)) .* wake.states.vortex_strengths[:,end]    

    end

end

function step_forward!(wake::ADMWake)
    #todo: not implemented
    X1 = Zygote.Buffer(wake.states.positions)
    G1 = Zygote.Buffer(wake.states.vortex_strengths)
    S1 = Zygote.Buffer(wake.states.core_sizes)
    p = wake.parameters
    q = wake.states
    X1[:,2:p.num_rings] = q.positions[:,1:p.num_rings-1] + 
        p.time_step * velocity(q.positions[:,1:p.num_rings-1], wake)
    G1[:,2:p.num_rings] = q.vortex_strengths[:,1:p.num_rings-1]
    # todo: implement vortex core size growth
    S1[:,2:p.num_rings] = q.core_sizes[:,1:p.num_rings-1]
    # l0 = []
    # S1[:,2:p.num_rings] = sqrt.(4*p.alpha*p.delta*p.kinematic_viscosity*p.time_step+ q.core_sizes[:,1:p.num_rings-1].^2) .* sqrt(l0./l1)
    
    # generate new ring
    X1[:,1] = initial_points(wake)
    x0 = q.positions[:,1:p.num_rings-1]
    x1 = X1[:,2:p.num_rings]
    
    l0 = norm.(x0[1:p.num_elements, :] - x0[(1:p.num_elements).%p.num_elements.+1,:])
    l1 = norm.(x1[1:p.num_elements, :] - x1[(1:p.num_elements).%p.num_elements.+1,:])
    S1[:,2:p.num_rings] = sqrt.(4*p.alpha*p.delta*p.kinematic_viscosity*p.time_step .+ q.core_sizes[:,1:p.num_rings-1].^2)  #.* sqrt.(l0./l1)

    # uniform distribution of vortex strength
    gamma = p.time_step * wake.controls.thrust / (wake.parameters.rotor.parameters.air_density * p.rotor.parameters.area)
    G1[:,1] = gamma * ones(p.num_elements)
    # uniform core size initialisation
    S1[:,1] = p.initial_core_size * ones(p.num_elements)

    wake.states.positions = copy(X1)
    wake.states.vortex_strengths = copy(G1)
    wake.states.core_sizes = copy(S1)
end

function step_forward_substep!(wake::ADMWake,substep=4)
    #todo: not implemented
    X1 = Zygote.Buffer(wake.states.positions)
    G1 = Zygote.Buffer(wake.states.vortex_strengths)
    S1 = Zygote.Buffer(wake.states.core_sizes)
    p = wake.parameters
    q = wake.states
    # X1[:,2:p.num_rings] = q.positions[:,1:p.num_rings-1] + 
    #     (p.time_step/substep) *  velocity(q.positions[:,1:p.num_rings-1], wake)
    X1[:,2:p.num_rings] = q.positions[:,1:p.num_rings-1]+ 
    (p.time_step/substep) * velocity(q.positions[:,1:p.num_rings-1], wake)
    X1[:,1] = initial_points(wake)
    
    G1[:,2:p.num_rings] = q.vortex_strengths[:,1:p.num_rings-1]
    # todo: implement vortex core size growth
    S1[:,2:p.num_rings] = q.core_sizes[:,1:p.num_rings-1]
    
    # generate new ring
    # uniform distribution of vortex strength
    gamma = (1. /substep) *p.time_step * wake.controls.thrust / (wake.parameters.rotor.parameters.air_density * p.rotor.parameters.area)
    G1[:,1] = gamma * ones(p.num_elements)
    # uniform core size initialisation
    S1[:,1] = p.initial_core_size * ones(p.num_elements)

    wake.states.positions = copy(X1)
    wake.states.vortex_strengths = copy(G1)
    wake.states.core_sizes = copy(S1)

    for step in 2:1:substep
        X1 = Zygote.Buffer(wake.states.positions)
        G1 = Zygote.Buffer(wake.states.vortex_strengths)
        S1 = Zygote.Buffer(wake.states.core_sizes)

        # X1[:,2:p.num_rings] = q.positions[:,2:p.num_rings] + 
        #  (p.time_step/substep)* velocity(q.positions[:,2:p.num_rings], wake)
        X1[:,2:p.num_rings] = q.positions[:,2:p.num_rings]+ 
         (p.time_step/substep) * velocity(q.positions[:,2:p.num_rings], wake)
        X1[:,1] = initial_points(wake)
        w2 = deepcopy(wake)
      
        G1[:,2:p.num_rings] = wake.states.vortex_strengths[:,2:p.num_rings]
        # todo: implement vortex core size growth
        S1[:] = wake.states.core_sizes
        
        # generate new ring
        X1[:,1] = initial_points(wake)
        # uniform distribution of vortex strength
        gamma = (step/substep) *p.time_step * wake.controls.thrust / (wake.parameters.rotor.parameters.air_density * p.rotor.parameters.area)
        G1[:,1] = gamma * ones(p.num_elements)
        # uniform core size initialisation
        # S1[:,1] = p.initial_core_size * ones(p.num_elements)
    
        wake.states.positions = copy(X1)
        wake.states.vortex_strengths = copy(G1)
        wake.states.core_sizes = copy(S1)
    
    end

end
function step_forward_Heun!(wake::ADMWake, substep=1)
 #todo: not implemented
 X1 = Zygote.Buffer(wake.states.positions)
 G1 = Zygote.Buffer(wake.states.vortex_strengths)
 S1 = Zygote.Buffer(wake.states.core_sizes)
 p = wake.parameters
 q = wake.states

 wake.states.vortex_strengths[:,end] .= 0
 # estimate state update with forward Euler
 velocity_k = velocity(q.positions[:,1:p.num_rings-1] , wake)
 X1_p = q.positions[:,1:p.num_rings-1] + (p.time_step/substep) * velocity_k
 
 
 w2 = deepcopy(wake)
 w2.states.positions[:,1] = initial_points(w2)
 w2.states.positions[:,2:end] = X1_p

 gamma = (p.time_step/substep) * wake.controls.thrust / (wake.parameters.rotor.parameters.air_density * p.rotor.parameters.area)
 w2.states.vortex_strengths[:,2:end] = w2.states.vortex_strengths[:,1:end-1]
 w2.states.vortex_strengths[:,1] = gamma* ones(p.num_elements)
 

 # w2.states.vortex_strengths[:,end] = (1-1/substep) .* w2.states.vortex_strengths[:,end]    
 # w2.states.vortex_strengths[2,1] = gamma
 w2.states.core_sizes[:,2:end] = w2.states.core_sizes[:,1:end-1]
 w2.states.core_sizes[:,1] .= p.initial_core_size
 velocity_k1 = velocity(X1_p , w2)

 X1[:,2:p.num_rings] = (1/2) *(p.time_step/substep)* (velocity_k .+ velocity_k1) + 
 q.positions[:,1:p.num_rings-1]
 G1[:,2:p.num_rings] = q.vortex_strengths[:,1:p.num_rings-1]
 # todo: implement vortex core size growth
 S1[:,2:p.num_rings] = q.core_sizes[:,1:p.num_rings-1]
 
 # generate new ring
 X1[:,1] = initial_points(wake)
 # uniform distribution of vortex strength
 gamma = (p.time_step/substep) * wake.controls.thrust / (wake.parameters.rotor.parameters.air_density * p.rotor.parameters.area)
 G1[:,1] =  gamma* ones(p.num_elements)
 # uniform core size initialisation
 S1[:,1] = p.initial_core_size * ones(p.num_elements)

 wake.states.positions = copy(X1)
 wake.states.vortex_strengths = copy(G1)
 wake.states.core_sizes = copy(S1)
 # wake.states.vortex_strengths[:,end] = (1-1/substep) .* wake.states.vortex_strengths[:,end]    

 for step in 2:1:substep
     X1 = Zygote.Buffer(wake.states.positions)
     G1 = Zygote.Buffer(wake.states.vortex_strengths)
     S1 = Zygote.Buffer(wake.states.core_sizes)
     p = wake.parameters
     q = wake.states
     velocity_k = velocity(q.positions[:,2:p.num_rings] , wake)
     X1_p = q.positions[:,2:p.num_rings] + (p.time_step/substep) * velocity_k
     
     w2 = deepcopy(wake)
     w2.states.positions[:,1] = initial_points(w2)
     w2.states.positions[:,2:end] = X1_p
     gamma = (step/substep) * p.time_step * wake.controls.thrust / (wake.parameters.rotor.parameters.air_density * p.rotor.parameters.area)
     w2.states.vortex_strengths[:,1] = gamma* ones(p.num_elements)

     w2.states.core_sizes[:,1] .= p.initial_core_size
     velocity_k1 = velocity(X1_p , w2)
 
     X1[:,2:p.num_rings] = (1/2) *(p.time_step/substep)* (velocity_k + velocity_k1) + 
     q.positions[:,2:p.num_rings]

     # generate new ring
     X1[:,1] = initial_points(wake)
     G1[:] = wake.states.vortex_strengths
     gamma = (step/substep) * p.time_step * wake.controls.thrust / (wake.parameters.rotor.parameters.air_density * p.rotor.parameters.area)

     G1[:,1] =  gamma* ones(p.num_elements)
     S1[:] = wake.states.core_sizes
     wake.states.positions = copy(X1)
     wake.states.vortex_strengths = copy(G1)
     wake.states.core_sizes = copy(S1)
     # wake.states.vortex_strengths[:,end] = ((1-step/substep)/(1-(step-1)/substep)) .* wake.states.vortex_strengths[:,end]    

 end

end

function step_forward!(wake::ADMRWake)
    #todo: not implemented
    X1 = Zygote.Buffer(wake.states.positions)
    G1 = Zygote.Buffer(wake.states.vortex_strengths)
    S1 = Zygote.Buffer(wake.states.core_sizes)
    p = wake.parameters
    q = wake.states
    X1[:,2:p.num_rings] = q.positions[:,1:p.num_rings-1] + 
        p.time_step * velocity(q.positions[:,1:p.num_rings-1], wake)
    G1[:,2:p.num_rings] = q.vortex_strengths[:,1:p.num_rings-1]
    # todo: implement vortex core size growth
    S1[:,2:p.num_rings] = q.core_sizes[:,1:p.num_rings-1]
    
    # generate new ring
    X1[:,1] = initial_points(wake)
    # x0 = q.positions[:,1:p.num_rings-1]
    # x1 = X1[:,2:p.num_rings]
    
    # l0 = norm.(x0[1:p.num_elements, :] - x0[(1:p.num_elements).%p.num_elements.+1,:])
    # l1 = norm.(x1[1:p.num_elements, :] - x1[(1:p.num_elements).%p.num_elements.+1,:])
    S1[1:p.num_elements,2:p.num_rings] = sqrt.(4*p.alpha*p.delta*p.kinematic_viscosity*p.time_step 
        .+ q.core_sizes[1:p.num_elements,1:p.num_rings-1].^2)  #.* sqrt.(l0./l1)

    # x0 = q.positions[:,:]
    # x1 = X1[:,:]
    # l0 = norm.(x0[p.num_elements+1, 1:p.num_rings-1] - x0[p.num_elements+1,2:p.num_rings])
    # l1 = norm.(x1[p.num_elements+1, 1:p.num_rings-1] - x1[p.num_elements+1,2:p.num_rings])
    S1[p.num_elements+1,2:p.num_rings] = sqrt.(4*p.alpha*p.delta*p.kinematic_viscosity*p.time_step 
        .+ q.core_sizes[p.num_elements+1,1:p.num_rings-1].^2)  #.* sqrt.(l0./l1)
    # uniform distribution of vortex strength
    gamma = p.time_step * wake.controls.thrust / (wake.parameters.rotor.parameters.air_density * p.rotor.parameters.area)
    G1[1:p.num_elements,1] = gamma * ones(p.num_elements)
    # G1[1:p.num_elements,1] = gamma .* 0.1 .* sin.(0.25 .* 2π.* range(start=0,step=2π/p.num_elements, length=p.num_elements))
    #todo: check root vortex strength!
    gamma_root = -2*wake.controls.thrust /  (p.rotor.parameters.air_density * wake.controls.omega * p.rotor.parameters.radius^2 )
    G1[p.num_elements+1,1] = gamma_root
    # uniform core size initialisation
    S1[:,1] = p.initial_core_size * ones(p.num_elements+1)

    wake.states.positions = copy(X1)
    wake.states.vortex_strengths = copy(G1)
    wake.states.core_sizes = copy(S1)
end

function step_forward!(wake::JoukowskyWake)
    #todo: not implemented
    X1 = Zygote.Buffer(wake.states.positions)
    G1 = Zygote.Buffer(wake.states.vortex_strengths)
    S1 = Zygote.Buffer(wake.states.core_sizes)
    p = wake.parameters
    q = wake.states
    X1[:,2:p.num_rings] = q.positions[:,1:p.num_rings-1] + 
        p.time_step * velocity(q.positions[:,1:p.num_rings-1], wake)
    # X1[p.num_blades+1,2:p.num_rings] = sum(X1[1:p.num_blades,2:p.num_rings],dims=1)./p.num_blades
    # move forward tip and root vorticity
    G1[:,2:p.num_rings] = q.vortex_strengths[:,1:p.num_rings-1]
    # G1[p.num_blades+1:2*p.num_blades+1,3:p.num_rings] = q.vortex_strengths[p.num_blades+1:2*p.num_blades+1,2:p.num_rings-1]
    # G1[p.num_blades+1:2*p.num_blades+1,2] = G1[1:p.num_blades+1,2:p.num_rings]


    # todo: implement vortex core size growth
    S1[:,2:p.num_rings] = q.core_sizes[:,1:p.num_rings-1]
    X1[:,1] = initial_points(wake)
    # x0 = q.positions[:,1:p.num_rings-1]
    # x1 = X1[:,2:p.num_rings]
    # # tip and root vortex core size
    x0 = q.positions[:,:]
    x1 = X1[:,:]
    l0 = norm.(x0[:, 1:p.num_rings-1] - x0[:,2:p.num_rings])
    l1 = norm.(x1[:, 1:p.num_rings-1] - x1[:,2:p.num_rings])
    S1[1:p.num_blades+1,2:p.num_rings] = sqrt.(4*p.alpha*p.delta*p.kinematic_viscosity*p.time_step 
        .+ q.core_sizes[1:p.num_blades+1,1:p.num_rings-1].^2) # .* sqrt.(l0./l1)

    # shed vortex core size
    x0 = q.positions[:,1:p.num_rings-1]
    x1 = X1[:,2:p.num_rings]
    l0 = norm.(x0[1:p.num_blades, :] - repeat(x0[p.num_blades+1:p.num_blades+1,:], p.num_blades))
    l1 = norm.(x1[1:p.num_blades, :] - repeat(x1[p.num_blades+1:p.num_blades+1,:], p.num_blades))
    S1[p.num_blades+1 .+ (1:p.num_blades),2:p.num_rings] = sqrt.(4*p.alpha*p.delta*p.kinematic_viscosity*p.time_step 
        .+ q.core_sizes[p.num_blades+1 .+ (1:p.num_blades),1:p.num_rings-1].^2)  #.* sqrt.(l0./l1)


    # uniform distribution of vortex strength
    #todo: refine into single function / joint with ADMRWake
    gamma_root =   -2*wake.controls.thrust /  (p.rotor.parameters.air_density * wake.controls.omega * p.rotor.parameters.radius^2 )
    # gamma_root =   2*2e6 /  (p.rotor.parameters.air_density * wake.controls.omega * p.rotor.parameters.radius^2 )
    gamma_bound = gamma_root / p.num_blades
    # tip vorticity
    G1[1:p.num_blades,1] = gamma_bound * ones(p.num_blades)
    # bound vorticity
    G1[p.num_blades+2:2*p.num_blades+1,1] =   gamma_bound * ones(p.num_blades)
    # shed vorticity
    # G1[p.num_blades+2:2*p.num_blades+1,2] = zeros(p.num_blades) #
    G1[p.num_blades+2:2*p.num_blades+1,2] = (q.vortex_strengths[p.num_blades+2:2*p.num_blades+1,1] -gamma_bound * ones(p.num_blades))
    #todo: check root vortex strength!
    G1[p.num_blades+1,1] = gamma_root
    
    # uniform core size initialisation
    S1[:,1] = p.initial_core_size * ones(2*p.num_blades+1)
    #(ones(2*p.num_blades+1) .+[0, 0, 0, 0.3*178.3, 0, 0, 0])

    wake.states.positions = copy(X1)
    wake.states.vortex_strengths = copy(G1)
    wake.states.core_sizes = copy(S1)
end