
function test_fixed_tower()
    base_position = @SVector randn(3)
    hubheight = rand()*100
    tilt = rand()*5
    # time_step = rand()

    params = FixedTowerParameters(base_position, hubheight, tilt) #, time_step)
    @test params.base_position == base_position
    @test params.hubheight == hubheight
    @test params.tilt == tilt
    # @test params.time_step == time_step

    thrust = randn()
    yaw = 20*randn()
    controls = FixedTowerControls(thrust, yaw)
    @test controls.thrust  ==  thrust
    @test controls.yaw == yaw

    states = FixedTowerStates()
    @test states.base_pitch == 0
    @test states.base_pitch_rate == 0
    @test states.base_surge == 0
    @test states.base_surge_rate == 0

    tower = FixedTower(params, states, controls)
    @test tower.parameters == params
    @test tower.controls == controls
    @test tower.states == states
    @test typeof(tower) == FixedTower
end

function test_default_fixed_tower()
    position =  @SVector randn(3)
    tower = construct_fixed_tower_10mw(position)
    @test typeof(tower) == FixedTower
    @test tower.parameters.base_position == position
    @test tower.parameters.hubheight == 119.
    @test tower.parameters.tilt == 0.
    @test tower.controls.thrust == 0
    @test tower.controls.yaw == 0
end

function test_update_controls()
    position =  @SVector randn(3)
    tower = construct_fixed_tower_10mw(position)
    new_thrust = rand()
    new_yaw = rand()
    update_controls!(tower, [new_thrust, new_yaw])
    @test tower.controls.thrust == new_thrust
    @test tower.controls.yaw == new_yaw
end

function test_rotation_matrices()
    theta = rand()
    @test isapprox(VortexWake.rot_x(theta)' * VortexWake.rot_x(theta), I)
    @test isapprox(VortexWake.rot_y(theta)' * VortexWake.rot_y(theta), I)
    @test isapprox(VortexWake.rot_z(theta)' * VortexWake.rot_z(theta), I)
    
    theta = rand()
    @test VortexWake.rot_x(theta)[:,1] == I(3)[:,1]
    @test VortexWake.rot_y(theta)[:,2] == I(3)[:,2]
    @test VortexWake.rot_z(theta)[:,3] == I(3)[:,3]

    @test isapprox(VortexWake.rot_x(π/2) * I(3)[:,3], -I(3)[:,2])    
    @test isapprox(VortexWake.rot_x(π) * I(3)[:,3], -I(3)[:,3])   

    @test isapprox(VortexWake.rot_y(π/2) * I(3)[:,3], I(3)[:,1])    
    @test isapprox(VortexWake.rot_y(π) * I(3)[:,3], -I(3)[:,3])   

    @test isapprox(VortexWake.rot_z(π/2) * I(3)[:,1], I(3)[:,2])    
    @test isapprox(VortexWake.rot_z(π) * I(3)[:,1], -I(3)[:,1])   
end
test_rotation_matrices()

function test_transform()
    points = [@SVector randn(3) for k in 1:1:10]
    tower = construct_fixed_tower_10mw(@SVector zeros(3))

    origin = @SVector zeros(3)
    origin_tf = VortexWake.transform_point(origin, tower)
    radius = 90
    top = SVector{3}([0,0,radius])
    side = SVector{3}([0,radius,0])
    top_tf = VortexWake.transform_point(top, tower)
    side_tf = VortexWake.transform_point(side, tower)
    @test isapprox(norm(top_tf-origin_tf), radius)
    @test isapprox(norm(side_tf-origin_tf), radius)

    point_tf = VortexWake.transform_point(points[1], tower)
    @test size(point_tf) == (3,)
    # @test distance

    points_tf = VortexWake.transform(points, tower)
    @test isapprox(point_tf, points_tf[1])

end
test_transform()
function test_motion()
    points = [@SVector randn(3) for k in 1:1:10]
    tower = construct_fixed_tower_10mw(@SVector zeros(3))
    point_velocity = motion(points, tower)
    for p in point_velocity
        @test isapprox(p, zeros(3))
    end
end

function test_rotor_direction()
    tower = construct_fixed_tower_10mw(@SVector zeros(3))
    n = rotor_direction(tower)
    @test size(n) == (3,)
end

test_fixed_tower()
test_default_fixed_tower()
test_update_controls()
test_rotation_matrices()
test_transform()
test_motion()
test_rotor_direction()