function test_rotor_points()
    p = rotor_disc_points()
end

function test_adm_wake()
    rotor = construct_momentum_rotor_10mw()
    tower = construct_fixed_tower_10mw(@SVector zeros(3))
    nr = rand(10:100)
    ne = rand(4:20)
    h = rand(0.1:0.01:1.)
    s0 = rand(0.1:0.01:1.)
    rotor_points = [@SVector zeros(3)] # todo: replace with rotor_points sampling
    params = ADMWakeParameters(rotor=rotor, tower=tower,
        num_rings=nr, num_elements=ne, rotor_points=rotor_points,
        time_step=h, initial_core_size=s0)

    @test params.rotor == rotor
    @test params.tower == tower
    @test params.num_rings == nr
    @test params.num_elements == ne
    @test params.rotor_points == rotor_points
    @test params.time_step == h
    @test params.initial_core_size == s0

    states = initialise_states(params)
    @test typeof(states) == ADMWakeStates

    # @show size(states.positions[])
    @test size(states.positions) == (params.num_elements, params.num_rings)
    @test size(states.vortex_strengths) == (params.num_elements, params.num_rings)
    @test size(states.core_sizes) == (params.num_elements, params.num_rings)

    # @test states.vortex_strengths
    # @test states.core_sizes 

    thrust = rand()
    inflow = @SVector rand(3)
    controls = ADMWakeControls(thrust, inflow)
    @test controls.thrust == thrust
    @test controls.inflow == inflow

    params = ADMRWakeParameters(rotor=rotor, tower=tower,
    num_rings=nr, num_elements=ne, rotor_points=rotor_points,
    time_step=h, initial_core_size=s0)

    @test params.rotor == rotor
    @test params.tower == tower
    @test params.num_rings == nr
    @test params.num_elements == ne
    @test params.rotor_points == rotor_points
    @test params.time_step == h
    @test params.initial_core_size == s0

    states = initialise_states(params)
    @test typeof(states) == ADMRWakeStates
    # @show size(states.positions[])
    @test size(states.positions) == (params.num_elements+1, params.num_rings)
    @test size(states.vortex_strengths) == (params.num_elements+1, params.num_rings)
    @test size(states.core_sizes) == (params.num_elements+1, params.num_rings)

    omega = rand()
    controls = ADMRWakeControls(thrust, inflow, omega)
    @test controls.thrust == thrust
    @test controls.inflow == inflow
    @test controls.omega == omega

    nb = rand(1:1:4)
    params = JoukowskyWakeParameters(rotor=rotor, tower=tower,
    num_rings=nr, num_blades=nb, rotor_points=rotor_points,
    time_step=h, initial_core_size=s0)

    @test params.rotor == rotor
    @test params.tower == tower
    @test params.num_rings == nr
    @test params.num_blades == nb
    @test params.rotor_points == rotor_points
    @test params.time_step == h
    @test params.initial_core_size == s0

    states = initialise_states(params)
    @test typeof(states) == JoukowskyWakeStates
    # @show size(states.positions[])
    @test size(states.positions) == (params.num_blades+1, params.num_rings)
    @test size(states.vortex_strengths) == (2*params.num_blades+1, params.num_rings)
    @test size(states.core_sizes) == (2*params.num_blades+1, params.num_rings)

    omega = rand()
    controls = JoukowskyWakeControls(thrust, inflow, omega)
    #todo: maybe change to rotating vs. non-rotating wake controls
    @test controls.thrust == thrust
    @test controls.inflow == inflow
    @test controls.omega == omega
end
test_adm_wake()




function test_initial_points()
    wake = construct_default_adm_wake()
    X0 = VortexWake.initial_points(wake)
    @test size(X0) == (wake.parameters.num_elements,)

    wake = construct_default_admr_wake()
    X0 = VortexWake.initial_points(wake)
    @test size(X0) == (wake.parameters.num_elements+1,)

    wake = construct_default_joukowsky_wake()
    X0 = VortexWake.initial_points(wake)
    @test size(X0) == (wake.parameters.num_blades+1,)
    #todo: test circular distribution
end
test_initial_points()

function test_induced_velocity()
    wake = construct_default_adm_wake()
    points = [@SVector randn(3) for k in 1:10]
    ui = induced_velocity(points, wake)
    for u in ui
        @test typeof(u) == typeof(points[1])
        @test u == zeros(3)
    end
    induced_velocity(wake.states.positions, wake)

    wake = construct_default_admr_wake()
    points = [@SVector randn(3) for k in 1:10]
    ui = induced_velocity(points, wake)
    for u in ui
        @test typeof(u) == typeof(points[1])
        @test u == zeros(3)
    end
    induced_velocity(wake.states.positions, wake)

    wake = construct_default_joukowsky_wake()
    points = [@SVector randn(3) for k in 1:10]
    ui = induced_velocity(points, wake)
    for u in ui
        @test typeof(u) == typeof(points[1])
        @test u == zeros(3)
    end
    induced_velocity(wake.states.positions, wake)
end
test_induced_velocity()

function test_rotor_velocity()
    wake = construct_default_adm_wake()
    u = rotor_velocity(wake)
    @test size(u) == ()
    @test u <= norm(wake.controls.inflow)

    wake = construct_default_admr_wake()
    u = rotor_velocity(wake)
    @test size(u) == ()
    @test u <= norm(wake.controls.inflow)

    wake = construct_default_joukowsky_wake()
    u = rotor_velocity(wake)
    @test size(u) == ()
    @test u <= norm(wake.controls.inflow)
end
test_rotor_velocity()

function test_update_controls()
    wake = construct_default_adm_wake()
    thrust = rand()
    inflow = @SVector randn(3)
    update_controls!(wake, [thrust, inflow...])
    @test wake.controls.thrust == thrust
    @test wake.controls.inflow == inflow

    wake = construct_default_admr_wake()
    thrust = rand()
    inflow = @SVector randn(3)
    omega = rand()
    update_controls!(wake, [thrust, inflow..., omega])
    @test wake.controls.thrust == thrust
    @test wake.controls.inflow == inflow
    @test wake.controls.omega == omega

    wake = construct_default_joukowsky_wake()
    thrust = rand()
    inflow = @SVector randn(3)
    omega = rand()
    update_controls!(wake, [thrust, inflow..., omega])
    @test wake.controls.thrust == thrust
    @test wake.controls.inflow == inflow
    @test wake.controls.omega == omega

end
test_update_controls()

function test_step_forward()
    wake = construct_default_adm_wake()
    step_forward!(wake)

    # [@test s== wake.parameters.initial_core_size for s in wake.states.core_sizes[:,1]]
    @test all(isapprox.(wake.states.core_sizes[:,1],wake.parameters.initial_core_size))

    wake = construct_default_admr_wake()
    step_forward!(wake)

    # [@test s== wake.parameters.initial_core_size for s in wake.states.core_sizes[:,1]]
    @test all(isapprox.(wake.states.core_sizes[:,1],wake.parameters.initial_core_size))

    wake = construct_default_joukowsky_wake()
    step_forward!(wake)

    # [@test s== wake.parameters.initial_core_size for s in wake.states.core_sizes[:,1]]
    @test all(isapprox.(wake.states.core_sizes[:,1],wake.parameters.initial_core_size))
    # do tip vortices sum to root vortex?
    q = wake.states
    @test all(isapprox.(sum(q.vortex_strengths[1:wake.parameters.num_blades,:],dims=1),
        q.vortex_strengths[wake.parameters.num_blades+1,:]))
    # does shed vorticity add up to tip vortex?
    # @test q.vortex_strengths[]
end
test_step_forward()

# test_rotor_points()
#
#test_gradients

function test_add_svector_to_matrix()
    X = [@SVector randn(3) for i1 in 1:1:5, i2 in 1:1:10]
    u = [SVector{3}(randn(3))]
    X_add = X.+u
    for idx in eachindex(X)
        @test X[idx]+u[1] == X_add[idx]
    end
end
test_add_svector_to_matrix()

function test_zygote_buffer_assignment()
    S0 = randn(4,5)
    S1 = Zygote.Buffer(S0)
    S1[:,2:5] = S0[:,1:4]
    S1[:,1] .= 2.
    S = copy(S1)
    # broadcast / element-wise assignment does not work
    @test ~all(isapprox.(S[:,1], 2 .* ones(4)))

    S1 = Zygote.Buffer(S0)
    S1[:,2:5] = S0[:,1:4]
    S1[:,1] = 2. * ones(4)
    S = copy(S1)
    @test all(isapprox.(S[:,1],2. * one(4)))
    # [@test isapprox(s,0) for s in S[:,2:5]-S0[:,1:5]]
    @test all(isapprox.(S[:,2:5],S0[:,1:4]))

end
test_zygote_buffer_assignment()

function test_vector_stack()
    a = [@SVector randn(3) for k  in 1:1:10]
    @test size(repeat(a', 3)) == (3,10)
end
test_vector_stack()

#test_vortex_strength()
#test_core_size_growth()

function test_boolean_filter()
    point = @SVector randn(3)
    X = [@SVector randn(3) for m in 1:10, n in 1:5]
    X1 = X[1:end-1,:]
    X2 = X[2:end,:]
    r0 = X2 .- X1
    r1 = X1 .- [point] #[xv - point for xv in X1] # X1 .- point
    r2 = X2 .- [point]
    threshold = 1.
    norm.(cross.(r1, r2)) .>=threshold .&& norm.(r1) .>= threshold
end
test_boolean_filter()